package org.fusesource.hawtbuf.proto;

option java_multiple_files = true;

message Enriched {
  enum Type {
    UNKNOWN = 0;
    FILE = 1;
    LINK = 2;
    COLLECTION = 3;
  }

  required Type type = 1;
  optional int64 size = 2;
  optional string mimeType = 3;
}

message SimpleMessage {
  required string nodeId = 1;
  required string repoId = 2;
  required string path = 3;
}

message EnrichedMessage {
  required string nodeId = 1;
  required string repoId = 2;
  required string path = 3;
  optional Enriched enriched = 4;
}
