// Protocol Buffers - Google's data interchange format
// Copyright 2008 Google Inc.
// http://code.google.com/p/protobuf/
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

// Author: kenton@google.com (Kenton Varda)
//  Based on original Protocol Buffers design by
//  Sanjay Ghemawat, Jeff Dean, and others.
//
// A proto file which imports a proto file that uses optimize_for = CODE_SIZE.

import "unittest_optimize_for.proto";

package protobuf_unittest;
option java_package = "org.fusesource.hawtbuf.proto";

// We optimize for speed here, but we are importing a proto that is optimized
// for code size.
option optimize_for = SPEED;

message TestEmbedOptimizedForSize {
  // Test that embedding a message which has optimize_for = CODE_SIZE into
  // one optimized for speed works.
  optional TestOptimizedForSize optional_message = 1;
  repeated TestOptimizedForSize repeated_message = 2;
}
