//
// Licensed to the Apache Software Foundation (ASF) under one or more
// contributor license agreements.  See the NOTICE file distributed with
// this work for additional information regarding copyright ownership.
// The ASF licenses this file to You under the Apache License, Version 2.0
// (the "License"); you may not use this file except in compliance with
// the License.  You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

package protobuf_unittest;
option java_package = "org.fusesource.hawtbuf.proto";
option java_outer_classname = "DeferredUnmarshal";
option deferred_decode = true;

message Foo {
 
  optional int32 field1 =  1;
  optional int64 field2 =  2;
 
}


message Bar {
  option base_type=Foo;

  // These are the Foo fields.  
  optional int32 field1 =  1;
  optional int64 field2 =  2;
 
  optional Foo field3 =  3;

}

